@@ -1,6 +1,6 @@
 //
 //  Little cms
-//  Copyright (C) 1998-2003 Marti Maria
+//  Copyright (C) 1998-2004 Marti Maria
 //
 // Permission is hereby granted, free of charge, to any person obtaining 
 // a copy of this software and associated documentation files (the "Software"), 
@@ -30,6 +30,11 @@
 #include <stdlib.h>
 #include <string.h>
 #include <stdarg.h>
+#include <sys/stat.h>
+
+#ifndef NON_WINDOWS
+#    include <io.h>
+#endif
 
 #include "jpeglib.h"
 #include "iccjpeg.h"
@@ -46,26 +51,37 @@ int    cdecl xgetopt(int argc, char *argv[], char *optionS);
 
 // ------------------------------------------------------------------------
 
+extern cmsHPROFILE OpenStockProfile(const char* File);
+
+
 // Flags
 
 static BOOL Verbose                = FALSE;
-static BOOL BlackWhiteCompensation = FALSE;
+static BOOL BlackPointCompensation = FALSE;
 static BOOL IgnoreEmbedded         = FALSE;
 static BOOL GamutCheck             = FALSE;
+static BOOL lIsITUFax              = FALSE;
+static BOOL lIsPhotoshopApp13      = FALSE;
+static BOOL lIsDeviceLink          = FALSE;
+static BOOL EmbedProfile           = FALSE;
+
+static const char* SaveEmbedded = NULL;
+
 static int Intent                  = INTENT_PERCEPTUAL;
 static int ProofingIntent          = INTENT_PERCEPTUAL;
 static int PrecalcMode             = 1;
-static BOOL lIsDeviceLink          = FALSE;
-static int jpegQuality = 75;
+
+static int jpegQuality             = 75;
 
 static char *cInpProf  = NULL;
 static char *cOutProf  = NULL;
 static char *cProofing = NULL;
+
 static FILE * InFile;
 static FILE * OutFile;
+
 static struct jpeg_decompress_struct Decompressor;
 static struct jpeg_compress_struct   Compressor;
-static BOOL   lIsITUFax = FALSE;
 
 
 static struct my_error_mgr {
@@ -78,17 +94,17 @@ static struct my_error_mgr {
 static
 void ConsoleWarningHandler(const char* module, const char* fmt, va_list ap)
 {
-        char e[512] = { '\0' };
-        if (module != NULL)
-              strcat(strcpy(e, module), ": ");
-
-        vsprintf(e+strlen(e), fmt, ap);
-        strcat(e, ".");
-        if (Verbose) {
-
-              fprintf(stderr, "\nWarning");
-              fprintf(stderr, " %s\n", e);
-              }
+    char e[512] = { '\0' };
+    if (module != NULL)
+        strcat(strcpy(e, module), ": ");
+    
+    vsprintf(e+strlen(e), fmt, ap);
+    strcat(e, ".");
+    if (Verbose) {
+        
+        fprintf(stderr, "\nWarning");
+        fprintf(stderr, " %s\n", e);
+    }
 }
 
 static
@@ -118,6 +134,13 @@ void FatalError(const char *frm, ...)
        exit(1);
 }
 
+static
+int MyErrorHandler(int ErrorCode, const char *ErrorText)
+{
+    FatalError("%s", ErrorText);
+    return 0;
+}
+
 
 // Out of mem
 static
@@ -153,60 +176,140 @@ static
 BOOL IsITUFax(jpeg_saved_marker_ptr ptr)
 {
     while (ptr) {
-
-    if (ptr -> marker == (JPEG_APP0 + 1) && ptr -> data_length > 5)
-      {
-        JOCTET FAR* data = ptr -> data;
-
-         if (GETJOCTET(data[0]) == 0x47 &&
-             GETJOCTET(data[1]) == 0x33 &&
-             GETJOCTET(data[2]) == 0x46 &&
-             GETJOCTET(data[3]) == 0x41 &&
-             GETJOCTET(data[4]) == 0x58)
-                    return TRUE;
-      }
-
-    ptr = ptr -> next;
-   }
+        
+        if (ptr -> marker == (JPEG_APP0 + 1) && ptr -> data_length > 5)
+        {
+            JOCTET FAR* data = ptr -> data;
+            
+            if (GETJOCTET(data[0]) == 0x47 &&
+                GETJOCTET(data[1]) == 0x33 &&
+                GETJOCTET(data[2]) == 0x46 &&
+                GETJOCTET(data[3]) == 0x41 &&
+                GETJOCTET(data[4]) == 0x58)
+                return TRUE;
+        }
+        
+        ptr = ptr -> next;
+    }
 
    return FALSE;
 
 }
 
+#define PS_FIXED_TO_FLOAT(h, l) ((float) (h) + ((float) (l)/(1<<16)))
+
+static
+BOOL ProcessPhotoshopAPP13(JOCTET FAR *data, int datalen)
+{
+    int i;
+
+    for (i = 14; i < datalen; )
+    {
+        long len;
+        unsigned int type;
+                
+        if (!(GETJOCTET(data[i]  ) == 0x38 &&
+              GETJOCTET(data[i+1]) == 0x42 &&
+              GETJOCTET(data[i+2]) == 0x49 &&
+              GETJOCTET(data[i+3]) == 0x4D)) break; // Not recognized
+                    
+        i += 4; // identifying string 
+        
+        type = (unsigned int) (GETJOCTET(data[i]<<8) + GETJOCTET(data[i+1]));
+
+        i += 2; // resource type 
+        
+        i += GETJOCTET(data[i]) + ((GETJOCTET(data[i]) & 1) ? 1 : 2);   // resource name 
+        
+        len = ((((GETJOCTET(data[i]<<8) + GETJOCTET(data[i+1]))<<8) + 
+                         GETJOCTET(data[i+2]))<<8) + GETJOCTET(data[i+3]);
+
+        i += 4; // Size 
+        
+        if (type == 0x03ED && len >= 16) {
+            
+            Decompressor.X_density = (int) PS_FIXED_TO_FLOAT(GETJOCTET(data[i]<<8) + GETJOCTET(data[i+1]),
+                                                 GETJOCTET(data[i+2]<<8) + GETJOCTET(data[i+3]));
+            Decompressor.Y_density = (int) PS_FIXED_TO_FLOAT(GETJOCTET(data[i+8]<<8) + GETJOCTET(data[i+9]),
+                                                 GETJOCTET(data[i+10]<<8) + GETJOCTET(data[i+11]));
+            
+            // Set the density unit to 1 since the 
+            // Vertical and Horizontal resolutions 
+            // are specified in Pixels per inch 
+
+            Decompressor.density_unit = 0x01;
+            return TRUE;
+            
+        }
+                
+        i += len + ((len & 1) ? 1 : 0);   // Alignment             
+    }
+    return FALSE;
+}
+
+
+static
+BOOL HandlePhotoshopAPP13(jpeg_saved_marker_ptr ptr)
+{
+    while (ptr) {
+        
+        if (ptr -> marker == (JPEG_APP0 + 13) && ptr -> data_length > 9)
+        {    
+            JOCTET FAR* data = ptr -> data;    
+            
+            if(GETJOCTET(data[0]) == 0x50 &&
+               GETJOCTET(data[1]) == 0x68 &&
+               GETJOCTET(data[2]) == 0x6F &&
+               GETJOCTET(data[3]) == 0x74 &&
+               GETJOCTET(data[4]) == 0x6F &&
+               GETJOCTET(data[5]) == 0x73 &&
+               GETJOCTET(data[6]) == 0x68 &&
+               GETJOCTET(data[7]) == 0x6F &&
+               GETJOCTET(data[8]) == 0x70) {
+             
+                ProcessPhotoshopAPP13(data, ptr -> data_length);
+                return TRUE;
+            }
+        }
+        
+        ptr = ptr -> next;
+    }
+    
+    return FALSE;    
+}
+
 
 
 static
 BOOL OpenInput(const char* FileName)
 {
-
     int m;
-
-        lIsITUFax = FALSE;   
-        InFile  = fopen(FileName, "rb");
-        if (InFile == NULL) {
-                    FatalError("Cannot open '%s'", FileName);
-                    return FALSE;
-        }
-
-        // Now we can initialize the JPEG decompression object.
-
-        Decompressor.err                 = jpeg_std_error(&ErrorHandler.pub);
-        ErrorHandler.pub.error_exit      = my_error_exit;
-        ErrorHandler.pub.output_message  = my_output_message;
-
-        jpeg_create_decompress(&Decompressor);
-        jpeg_stdio_src(&Decompressor, InFile);
-      
-        for (m = 0; m < 16; m++)
-            jpeg_save_markers(&Decompressor, JPEG_APP0 + m, 0xFFFF);
-
-        setup_read_icc_profile(&Decompressor);  
-
-         fseek(InFile, 0, SEEK_SET);
-        jpeg_read_header(&Decompressor, TRUE);
-        // jpeg_save_markers(&Decompressor, JPEG_APP0+1, 0xffff);   // ITU T.42 G3FAX
-       
-        return TRUE;
+    
+    lIsITUFax = FALSE;   
+    InFile  = fopen(FileName, "rb");
+    if (InFile == NULL) {
+        FatalError("Cannot open '%s'", FileName);
+        return FALSE;
+    }
+    
+    // Now we can initialize the JPEG decompression object.
+    
+    Decompressor.err                 = jpeg_std_error(&ErrorHandler.pub);
+    ErrorHandler.pub.error_exit      = my_error_exit;
+    ErrorHandler.pub.output_message  = my_output_message;
+    
+    jpeg_create_decompress(&Decompressor);
+    jpeg_stdio_src(&Decompressor, InFile);
+    
+    for (m = 0; m < 16; m++)
+        jpeg_save_markers(&Decompressor, JPEG_APP0 + m, 0xFFFF);
+    
+    setup_read_icc_profile(&Decompressor);  
+    
+    fseek(InFile, 0, SEEK_SET);
+    jpeg_read_header(&Decompressor, TRUE);
+    
+    return TRUE;
 }
 
 
@@ -241,21 +344,25 @@ BOOL Done()
        
 }
 
+
 // Build up the pixeltype descriptor
 
 static
 DWORD GetInputPixelType(void)
 {
-    
-     int space, bps, extra, ColorChannels;
+     int space, bps, extra, ColorChannels, Flavor;
         
-     lIsITUFax = IsITUFax(Decompressor.marker_list);
-     ColorChannels = Decompressor.num_components;
-     extra = 0;             // Alpha = None
-     bps   = 1;             // 8 bits
+   
+     lIsITUFax         = IsITUFax(Decompressor.marker_list);
+     lIsPhotoshopApp13 = HandlePhotoshopAPP13(Decompressor.marker_list);
 
+     ColorChannels = Decompressor.num_components;
+     extra  = 0;            // Alpha = None
+     bps    = 1;            // 8 bits
+     Flavor = 0;            // Vanilla
 
      if (lIsITUFax) {
+
         space = PT_Lab;
         Decompressor.out_color_space = JCS_YCbCr;  // Fake to don't touch
      }
@@ -280,11 +387,15 @@ DWORD GetInputPixelType(void)
      case JCS_CMYK:            // C/M/Y/K
               space = PT_CMYK;
               Decompressor.out_color_space = JCS_CMYK;
+              if (Decompressor.saw_Adobe_marker)            // Adobe keeps CMYK inverted, so change flavor
+                                Flavor = 1;                 // from vanilla to chocolate
               break;
 
      case JCS_YCCK:            // Y/Cb/Cr/K              
               space = PT_CMYK;
               Decompressor.out_color_space = JCS_CMYK;
+              if (Decompressor.saw_Adobe_marker)            // ditto
+                                Flavor = 1;                 
               break;
 
      default:
@@ -292,7 +403,7 @@ DWORD GetInputPixelType(void)
               return 0;
      }
 
-     return (EXTRA_SH(extra)|CHANNELS_SH(ColorChannels)|BYTES_SH(bps)|COLORSPACE_SH(space)|PLANAR_SH(FALSE));
+     return (EXTRA_SH(extra)|CHANNELS_SH(ColorChannels)|BYTES_SH(bps)|COLORSPACE_SH(space)|FLAVOR_SH(Flavor));
 }
 
 
@@ -305,7 +416,8 @@ static
 DWORD ComputeOutputFormatDescriptor(DWORD dwInput, int OutColorSpace)
 {
    int IsPlanar  = T_PLANAR(dwInput);
-   int Channels = 0;
+   int Channels  = 0;
+   int Flavor    = 0;
 
    switch (OutColorSpace) {
 
@@ -317,17 +429,20 @@ DWORD ComputeOutputFormatDescriptor(DWORD dwInput, int OutColorSpace)
    case PT_Lab:
    case PT_YUV:
    case PT_YCbCr:
-           Channels = 3;
+               Channels = 3;
                break;
 
    case PT_CMYK:
-           Channels = 4;
+               if (Compressor.write_Adobe_marker)   // Adobe keeps CMYK inverted, so change flavor to chocolate
+                        Flavor = 1;
+
+               Channels = 4;
                break;
    default:
                FatalError("Unsupported output color space");
    }
 
-    return (COLORSPACE_SH(OutColorSpace)|PLANAR_SH(IsPlanar)|CHANNELS_SH(Channels)|BYTES_SH(1));
+    return (COLORSPACE_SH(OutColorSpace)|PLANAR_SH(IsPlanar)|CHANNELS_SH(Channels)|BYTES_SH(1)|FLAVOR_SH(Flavor));
 }
 
 
@@ -355,6 +470,7 @@ int GetProfileColorSpace(cmsHPROFILE hProfile)
 
        case icSigHexachromeData: return PT_HiFi;
 
+
        default:  return icMaxEnumData;
        }
 }
@@ -381,6 +497,7 @@ void jcopy_markers_execute(j_decompress_ptr srcinfo, j_compress_ptr dstinfo)
     GETJOCTET(marker->data[3]) == 0x46 &&
     GETJOCTET(marker->data[4]) == 0)
       continue;         /* reject duplicate JFIF */
+
     if (dstinfo->write_Adobe_marker &&
     marker->marker == JPEG_APP0+14 &&
     marker->data_length >= 5 &&
@@ -390,13 +507,15 @@ void jcopy_markers_execute(j_decompress_ptr srcinfo, j_compress_ptr dstinfo)
     GETJOCTET(marker->data[3]) == 0x62 &&
     GETJOCTET(marker->data[4]) == 0x65)
       continue;         /* reject duplicate Adobe */
+
 #ifdef NEED_FAR_POINTERS
     /* We could use jpeg_write_marker if the data weren't FAR... */
     {
       unsigned int i;
+
       jpeg_write_m_header(dstinfo, marker->marker, marker->data_length);
       for (i = 0; i < marker->data_length; i++)
-    jpeg_write_m_byte(dstinfo, marker->data[i]);
+        jpeg_write_m_byte(dstinfo, marker->data[i]);
     }
 #else
     jpeg_write_marker(dstinfo, marker->marker,
@@ -405,47 +524,133 @@ void jcopy_markers_execute(j_decompress_ptr srcinfo, j_compress_ptr dstinfo)
   }
 }
 
-
 static
 void WriteOutputFields(int OutputColorSpace)
 {
-    int n, m, k;
+    J_COLOR_SPACE in_space, jpeg_space;
+    int components;
     
     switch (OutputColorSpace) {
         
-    case PT_GRAY: n = m = JCS_GRAYSCALE; k = 1; break;
-    case PT_RGB:  n = JCS_RGB; m = JCS_YCbCr; k = 3; break;       // red/green/blue
-    case PT_YCbCr: n = m = JCS_YCbCr; k = 3; break;               // Y/Cb/Cr (also known as YUV)
-    case PT_CMYK: n = m = JCS_CMYK;  k = 4; break;                // C/M/Y/K              
-    case PT_Lab:  n = m = JCS_YCbCr; k = 3; break;                // Fake to don't touch
+    case PT_GRAY: in_space = jpeg_space = JCS_GRAYSCALE; 
+                  components = 1; 
+                  break;
+
+    case PT_RGB:  in_space = JCS_RGB; 
+                  jpeg_space = JCS_YCbCr; 
+                  components = 3; 
+                  break;       // red/green/blue
+
+    case PT_YCbCr: in_space = jpeg_space = JCS_YCbCr; 
+                   components = 3; 
+                   break;               // Y/Cb/Cr (also known as YUV)
+
+    case PT_CMYK: in_space = JCS_CMYK; 
+                  jpeg_space = JCS_YCCK;  
+                  components = 4; 
+                  break;      // C/M/Y/components              
+
+    case PT_Lab:  in_space = jpeg_space = JCS_YCbCr; 
+                  components = 3; 
+                  break;                // Fake to don't touch
     default:
-        FatalError("Unsupported output color space");
-        return;
+                FatalError("Unsupported output color space");
+                return;
     }
     
-    Compressor.in_color_space = n;
-    Compressor.jpeg_color_space = m;
-    Compressor.input_components = Compressor.num_components = k;
+
+    if (jpegQuality >= 100) {
+
+     // avoid destructive conversion when asking for lossless compression 
+        jpeg_space = in_space;
+    }
+
+    Compressor.in_color_space =  in_space;
+    Compressor.jpeg_color_space = jpeg_space;
+    Compressor.input_components = Compressor.num_components = components;
     jpeg_set_defaults(&Compressor);
-    jpeg_set_colorspace(&Compressor, m);
-    jpeg_set_quality(&Compressor, jpegQuality,1);    
+    jpeg_set_colorspace(&Compressor, jpeg_space);
+    jpeg_set_quality(&Compressor, jpegQuality, 1);    
+
+    //avoid subsampling on high quality factor
+
+    if (jpegQuality >= 70) {
+
+      int i;
+      for(i=0; i < Compressor.num_components; i++) {
+
+	        Compressor.comp_info[i].h_samp_factor = 1;
+            Compressor.comp_info[i].v_samp_factor = 1;
+      }
+
+    }
+
 }
 
+// A replacement for (the nonstandard) filelenght
+
+static
+int xfilelength(int fd)
+{
+#ifdef _MSC_VER
+        return _filelength(fd);
+#else
+        struct stat sb;
+        if (fstat(fd, &sb) < 0)
+                return(-1);
+        return(sb.st_size);
+#endif
+
+
+}
+
+static
+void DoEmbedProfile(const char* ProfileFile)
+{
+    FILE* f;
+    size_t size, EmbedLen;
+    LPBYTE EmbedBuffer;
+
+        f = fopen(ProfileFile, "rb");
+        if (f == NULL) return;
+
+        size = xfilelength(fileno(f));
+        EmbedBuffer = (LPBYTE) malloc(size + 1);
+        EmbedLen = fread(EmbedBuffer, 1, size, f);
+        fclose(f);
+        EmbedBuffer[EmbedLen] = 0;
+
+        write_icc_profile (&Compressor, EmbedBuffer, EmbedLen);		   
+        free(EmbedBuffer);
+}
+
+
+
 static
 int DoTransform(cmsHTRANSFORM hXForm)
 {       
     JSAMPROW ScanLineIn;
     JSAMPROW ScanLineOut;
 
+    
+       //Preserve resolution values from the original
+       // (Thanks to robert bergs for finding out this bug)
 
-      
+       Compressor.density_unit = Decompressor.density_unit;
+       Compressor.X_density    = Decompressor.X_density;
+       Compressor.Y_density    = Decompressor.Y_density;
+
+      //  Compressor.write_JFIF_header = 1;
+     
        jpeg_start_decompress(&Decompressor);
        jpeg_start_compress(&Compressor, TRUE);
 
-      // FIXME: # of bytes is *4 at most, but can be less
+       // Embed the profile if needed
+       if (EmbedProfile && cOutProf) 
+           DoEmbedProfile(cOutProf);
 
-       ScanLineIn  = malloc(Decompressor.output_width * 4);
-       ScanLineOut = malloc(Compressor.image_width * 4);
+       ScanLineIn  = (JSAMPROW) malloc(Decompressor.output_width * Decompressor.num_components);
+       ScanLineOut = (JSAMPROW) malloc(Compressor.image_width * Compressor.num_components);
 
        while (Decompressor.output_scanline <
                             Decompressor.output_height) {
@@ -460,6 +665,7 @@ int DoTransform(cmsHTRANSFORM hXForm)
 
        free(ScanLineIn); 
        free(ScanLineOut);
+
        jpeg_finish_decompress(&Decompressor);
        jpeg_finish_compress(&Compressor);
        
@@ -469,7 +675,19 @@ int DoTransform(cmsHTRANSFORM hXForm)
 
 
 
+static
+void SaveMemoryBlock(const BYTE* Buffer, DWORD dwLen, const char* Filename)
+{
+    FILE* out = fopen(Filename, "wb");
+    if (out == NULL)
+           FatalError("Cannot create '%s'", Filename);
 
+    if (fwrite(Buffer, 1, dwLen, out) != dwLen)
+            FatalError("Cannot write %ld bytes to %s", dwLen, Filename);
+
+    if (fclose(out) != 0) 
+            FatalError("Error flushing file '%s'", Filename);
+}
 
 
 
@@ -487,11 +705,9 @@ int TransformImage(char *cDefInpProf, char *cOutProf)
        LPBYTE EmbedBuffer;
 
 
-       if (BlackWhiteCompensation) {
+       if (BlackPointCompensation) {
 
-            dwFlags |= cmsFLAGS_WHITEBLACKCOMPENSATION;
-            if (PrecalcMode == 0) 
-                    FatalError("Cannot use White/Black compensation without precalculation. Use /c1 or /c2.");
+            dwFlags |= cmsFLAGS_BLACKPOINTCOMPENSATION;            
        }
 
        switch (PrecalcMode) {
@@ -519,25 +735,23 @@ int TransformImage(char *cDefInpProf, char *cOutProf)
         {
               hIn = cmsOpenProfileFromMem(EmbedBuffer, EmbedLen);
               if (Verbose) fprintf(stdout, " (embedded profile found)");
+               if (hIn != NULL && SaveEmbedded != NULL)
+                          SaveMemoryBlock(EmbedBuffer, EmbedLen, SaveEmbedded);
+
+              free(EmbedBuffer);
         }
         else
         {
-              if (cDefInpProf == NULL)
-                  hIn = cmsCreate_sRGBProfile();
-              else
-                hIn = cmsOpenProfileFromFile(cDefInpProf, "r");
+                hIn = OpenStockProfile(cDefInpProf);
        }
 
-       if (cOutProf == NULL)
-        hOut = cmsCreate_sRGBProfile();
-       else
-        hOut = cmsOpenProfileFromFile(cOutProf, "r");
+        hOut = OpenStockProfile(cOutProf);
 
 
        hProof = NULL;
        if (cProofing != NULL) {
 
-           hProof = cmsOpenProfileFromFile(cProofing, "r");
+           hProof = OpenStockProfile(cProofing);
        }
         }
 
@@ -561,18 +775,18 @@ int TransformImage(char *cDefInpProf, char *cOutProf)
             OutputColorSpace = GetProfileColorSpace(hOut);
         }
 
-       wOutput      = ComputeOutputFormatDescriptor(wInput, OutputColorSpace);
-       
        jpeg_copy_critical_parameters(&Decompressor, &Compressor);
        
        WriteOutputFields(OutputColorSpace);               
        
+       wOutput      = ComputeOutputFormatDescriptor(wInput, OutputColorSpace);
+       
        xform = cmsCreateProofingTransform(hIn, wInput, 
                                           hOut, wOutput, 
                                           hProof, Intent, 
                                           ProofingIntent, dwFlags);
 
-       // Handle tile by tile or strip by strip
+       // Handle tile by tile or strip by strip strtok
 
        DoTransform(xform);
 
@@ -593,7 +807,7 @@ int TransformImage(char *cDefInpProf, char *cOutProf)
 static
 void Help(int level)
 {
-     fprintf(stderr, "little cms ICC profile applier for JPEG - v1.4\n\n");
+     fprintf(stderr, "little cms ICC profile applier for JPEG - v2.0\n\n");
 
      switch(level) {
 
@@ -607,17 +821,21 @@ void Help(int level)
      fprintf(stderr, "%ci<profile> - Input profile (defaults to sRGB)\n", SW);
      fprintf(stderr, "%co<profile> - Output profile (defaults to sRGB)\n", SW);   
      fprintf(stderr, "%ct<0,1,2,3> - Intent (0=Perceptual, 1=Colorimetric, 2=Saturation, 3=Absolute)\n", SW);    
-     fprintf(stderr, "%cm<0,1,2,3> - SoftProof intent\n", SW);
+     
      fprintf(stderr, "\n");
 
-     fprintf(stderr, "%cb - Black/White compensation\n", SW);
+     fprintf(stderr, "%cb - Black point compensation\n", SW);
      fprintf(stderr, "%cn - Ignore embedded profile\n", SW);
+     fprintf(stderr, "%ce - Embed destination profile\n", SW);
+     fprintf(stderr, "%cs<new profile> - Save embedded profile as <new profile>\n", SW);
 
+     fprintf(stderr, "\n");
 
      fprintf(stderr, "%cc<0,1,2,3> - Precalculates transform (0=Off, 1=Normal, 2=Hi-res, 3=LoRes) [defaults to 1]\n", SW);     
      fprintf(stderr, "\n");
 
      fprintf(stderr, "%cp<profile> - Soft proof profile\n", SW);
+     fprintf(stderr, "%cm<0,1,2,3> - SoftProof intent\n", SW);
      fprintf(stderr, "%cg - Marks out-of-gamut colors on softproof\n", SW);
 
      fprintf(stderr, "\n");
@@ -663,105 +881,113 @@ void Help(int level)
 static
 void HandleSwitches(int argc, char *argv[])
 {
-       int s;
-      
-       while ((s=xgetopt(argc,argv,"bBnNvVGgh:H:i:I:o:O:P:p:t:T:c:C:Q:q:M:m:L:l:")) != EOF) {
-
-       switch (s)
-       {
-
-       case 'b':
-       case 'B':
-            BlackWhiteCompensation = TRUE;
+    int s;
+    
+    while ((s=xgetopt(argc,argv,"bBnNvVGgh:H:i:I:o:O:P:p:t:T:c:C:Q:q:M:m:L:l:eEs:S:")) != EOF) {
+        
+        switch (s)
+        {
+            
+        case 'b':
+        case 'B':
+            BlackPointCompensation = TRUE;
             break;
-
-       case 'v':
-       case 'V':
+            
+        case 'v':
+        case 'V':
             Verbose = TRUE;
             break;
-
-       case 'i':
-       case 'I':
-
+            
+        case 'i':
+        case 'I':
+            
             if (lIsDeviceLink)
-                   FatalError("Device-link already specified"); 
-
+                FatalError("Device-link already specified"); 
+            
             cInpProf = xoptarg;
             break;
-
-       case 'o':
-       case 'O':
-
-             if (lIsDeviceLink)
-                   FatalError("Device-link already specified"); 
-
+            
+        case 'o':
+        case 'O':
+            
+            if (lIsDeviceLink)
+                FatalError("Device-link already specified"); 
+            
             cOutProf = xoptarg;
             break;
-
-       case 'l':
-       case 'L': 
-                cInpProf = xoptarg;
-                lIsDeviceLink = TRUE;
-                break;
-
-       case 'p':
-       case 'P':
-           cProofing = xoptarg;
-           break;
-
-       case 't':
-       case 'T':
+            
+        case 'l':
+        case 'L': 
+            cInpProf = xoptarg;
+            lIsDeviceLink = TRUE;
+            break;
+            
+        case 'p':
+        case 'P':
+            cProofing = xoptarg;
+            break;
+            
+        case 't':
+        case 'T':
             Intent = atoi(xoptarg);
             if (Intent > 3) Intent = 3;
             if (Intent < 0) Intent = 0;
             break;
-
-       case 'N':
-       case 'n':
+            
+        case 'N':
+        case 'n':
             IgnoreEmbedded = TRUE;
             break;
-
-       
+            
+        case 'e':
+        case 'E':
+            EmbedProfile = TRUE;
+            break;
+            
         case 'g':
         case 'G':
             GamutCheck = TRUE;
             break;
-
+            
         case 'c':
         case 'C':
             PrecalcMode = atoi(xoptarg);
             if (PrecalcMode < 0 || PrecalcMode > 2)
-                    FatalError("Unknown precalc mode '%d'", PrecalcMode);
+                FatalError("Unknown precalc mode '%d'", PrecalcMode);
             break;
-
+            
         case 'H':
         case 'h':  {
-
+            
             int a =  atoi(xoptarg);
             Help(a); 
-            }
+                   }
             break;
-
+            
         case 'q':
         case 'Q':
             jpegQuality = atoi(xoptarg);
             if (jpegQuality > 100) jpegQuality = 100;
             if (jpegQuality < 0)   jpegQuality = 0;
             break;
-
-       case 'm':
-       case 'M':
+            
+        case 'm':
+        case 'M':
             ProofingIntent = atoi(xoptarg);
             if (ProofingIntent > 3) ProofingIntent = 3;
             if (ProofingIntent < 0) ProofingIntent = 0;
             break;
-
-
-  default:
-
-       FatalError("Unknown option - run without args to see valid ones");
-    }
-       
+            
+        case 's':
+        case 'S': SaveEmbedded = xoptarg;
+            break;
+            
+            
+        default:
+            
+            FatalError("Unknown option - run without args to see valid ones");
+        }
+        
     }
 }
 
@@ -777,6 +1003,8 @@ int main(int argc, char* argv[])
 
       HandleSwitches(argc, argv);
 
+      cmsSetErrorHandler(MyErrorHandler);
+
       if ((argc - xoptind) != 2) {
 
               Help(0);              
