@@ -1,6 +1,6 @@
 //
 //  Little cms
-//  Copyright (C) 1998-2003 Marti Maria
+//  Copyright (C) 1998-2004 Marti Maria
 //
 // Permission is hereby granted, free of charge, to any person obtaining 
 // a copy of this software and associated documentation files (the "Software"), 
@@ -121,7 +121,10 @@ void NC2toPCS(struct _cmstransform_struct *p,
 {
     int index = In[0];
 
-    CopyMemory(Out, p ->NamedColorList->List[index].PCS, 3 * sizeof(WORD));
+    if (index >= p ->NamedColorList-> nColors)
+        cmsSignalError(LCMS_ERRC_WARNING, "Color %d out of range", index);
+    else
+        CopyMemory(Out, p ->NamedColorList->List[index].PCS, 3 * sizeof(WORD));
 }
 
 // From PCS to Shaper-Matrix
@@ -226,20 +229,14 @@ void NormalXFORM(_LPcmsTRANSFORM p,
               COPY_3CHANS(wPCS, wStageABC);
 
 
-       if (p->Gamut)
-       {
+       if (p->Gamut) {
 
        // Gamut check, enabled across CLUT
 
        cmsEvalLUT(p -> Gamut, wPCS, wGamut);
        
-#ifdef DEBUG         
-         printf("Gamut %d\n", wGamut[0]);
-#endif
-
-       if (wGamut[0] >= 1) {              // 8 for roundoff 
-        
-              
+       if (wGamut[0] >= 1) {              
+                      
               wOut[0] = AlarmR;          // Gamut alarm
               wOut[1] = AlarmG;
               wOut[2] = AlarmB;
@@ -259,7 +256,9 @@ void NormalXFORM(_LPcmsTRANSFORM p,
        }
 
        if (p -> Stage2) {
+
               p -> Stage2(wPCS, wStageLMN, &p->m2, &p->of2);
+           
               if (wPCS[0] == 0xFFFF &&
                   wPCS[1] == 0xFFFF &&
                   wPCS[2] == 0xFFFF) {
@@ -302,14 +301,153 @@ void PrecalculatedXFORM(_LPcmsTRANSFORM p,
        output = (LPBYTE) out;
        n = Size;                    // Buffer len
 
-       for (i=0; i < n; i++)
-       {
+       for (i=0; i < n; i++) {
+
        accum = p -> FromInput(p, wIn, accum);
        cmsEvalLUT(p -> DeviceLink, wIn, wOut);
        output = p -> ToOutput(p, wOut, output);
        }
 }
 
+// Auxiliar: Handle precalculated gamut check
+
+static
+void TransformOnePixelWithGamutCheck(_LPcmsTRANSFORM p, WORD wIn[], WORD wOut[])
+{
+    WORD wOutOfGamut;
+
+       cmsEvalLUT(p ->GamutCheck,  wIn, &wOutOfGamut);
+
+       if (wOutOfGamut >= 1) {
+
+              ZeroMemory(wOut, sizeof(WORD) * MAXCHANNELS);
+
+              wOut[0] = AlarmR; 
+              wOut[1] = AlarmG; 
+              wOut[2] = AlarmB; 
+              
+       }
+       else
+            cmsEvalLUT(p -> DeviceLink, wIn, wOut);
+
+}
+
+
+static
+void PrecalculatedXFORMGamutCheck(_LPcmsTRANSFORM p,
+                                  LPVOID in,
+                                  LPVOID out, unsigned int Size)
+{
+       register LPBYTE accum;
+       register LPBYTE output;
+       WORD wIn[MAXCHANNELS], wOut[MAXCHANNELS];
+       register unsigned int i, n;
+
+
+       accum  = (LPBYTE) in;
+       output = (LPBYTE) out;
+       n = Size;                    // Buffer len
+
+       for (i=0; i < n; i++) {
+
+       accum = p -> FromInput(p, wIn, accum);
+
+       TransformOnePixelWithGamutCheck(p, wIn, wOut);
+
+       output = p -> ToOutput(p, wOut, output);
+       }
+}
+
+
+
+// Using precalculated LUT + Cache
+
+static
+void CachedXFORM(_LPcmsTRANSFORM p,
+                     LPVOID in,
+                     LPVOID out, unsigned int Size)
+{
+       register LPBYTE accum;
+       register LPBYTE output;
+       WORD wIn[MAXCHANNELS], wOut[MAXCHANNELS];
+       register unsigned int i, n;
+
+
+       accum  = (LPBYTE) in;
+       output = (LPBYTE) out;
+       n = Size;                    // Buffer len
+
+       // Empty buffers for quick memcmp
+
+       ZeroMemory(wIn,  sizeof(WORD) * MAXCHANNELS);
+       ZeroMemory(wOut, sizeof(WORD) * MAXCHANNELS);
+
+       for (i=0; i < n; i++) {
+
+       accum = p -> FromInput(p, wIn, accum);
+       
+       if (memcmp(wIn, p ->CacheIn, sizeof(WORD) * MAXCHANNELS) == 0) {
+
+            CopyMemory(wOut, p ->CacheOut, sizeof(WORD) * MAXCHANNELS);
+       }
+       else {
+
+            CopyMemory(p ->CacheIn, wIn, sizeof(WORD) * MAXCHANNELS);
+
+            cmsEvalLUT(p -> DeviceLink, wIn, wOut);
+            
+            CopyMemory(p ->CacheOut, wOut, sizeof(WORD) * MAXCHANNELS);
+       }
+
+       output = p -> ToOutput(p, wOut, output);
+       }
+}
+
+
+
+// Using precalculated LUT + Cache
+
+static
+void CachedXFORMGamutCheck(_LPcmsTRANSFORM p,
+                           LPVOID in,
+                           LPVOID out, unsigned int Size)
+{
+       register LPBYTE accum;
+       register LPBYTE output;
+       WORD wIn[MAXCHANNELS], wOut[MAXCHANNELS];
+       register unsigned int i, n;
+
+
+       accum  = (LPBYTE) in;
+       output = (LPBYTE) out;
+       n = Size;                    // Buffer len
+
+       // Empty buffers for quick memcmp
+
+       ZeroMemory(wIn,  sizeof(WORD) * MAXCHANNELS);
+       ZeroMemory(wOut, sizeof(WORD) * MAXCHANNELS);
+
+       for (i=0; i < n; i++) {
+
+       accum = p -> FromInput(p, wIn, accum);
+       
+       if (memcmp(wIn, p ->CacheIn, sizeof(WORD) * MAXCHANNELS) == 0) {
+
+            CopyMemory(wOut, p ->CacheOut, sizeof(WORD) * MAXCHANNELS);
+       }
+       else {
+
+            CopyMemory(p ->CacheIn, wIn, sizeof(WORD) * MAXCHANNELS);
+
+            TransformOnePixelWithGamutCheck(p, wIn, wOut);
+            
+            CopyMemory(p ->CacheOut, wOut, sizeof(WORD) * MAXCHANNELS);
+       }
+
+       output = p -> ToOutput(p, wOut, output);
+       }
+}
+
 
 // Using smelted Matrix/Shaper
 
@@ -441,7 +579,7 @@ LPMATSHAPER cmsBuildGrayOutputMatrixShaper(cmsHPROFILE hProfile)
 
 // Input matrix, only in XYZ
 
-LPMATSHAPER cmsBuildInputMatrixShaper(cmsHPROFILE InputProfile, LPDWORD dwFlags)
+LPMATSHAPER cmsBuildInputMatrixShaper(cmsHPROFILE InputProfile)
 {
        MAT3 DoubleMat;
        LPGAMMATABLE Shapes[3];
@@ -453,7 +591,7 @@ LPMATSHAPER cmsBuildInputMatrixShaper(cmsHPROFILE InputProfile, LPDWORD dwFlags)
 
        if (cmsGetColorSpace(InputProfile) == icSigGrayData)
        {
-              if (dwFlags) *dwFlags |= cmsFLAGS_NOTPRECALC;
+             // if (dwFlags) *dwFlags |= cmsFLAGS_NOTPRECALC;
               return cmsBuildGrayInputMatrixShaper(InputProfile);              
        }
 
@@ -478,7 +616,7 @@ LPMATSHAPER cmsBuildInputMatrixShaper(cmsHPROFILE InputProfile, LPDWORD dwFlags)
 // Output style matrix-shaper
 
 
-LPMATSHAPER cmsBuildOutputMatrixShaper(cmsHPROFILE OutputProfile, LPDWORD dwFlags)
+LPMATSHAPER cmsBuildOutputMatrixShaper(cmsHPROFILE OutputProfile)
 {
        MAT3 DoubleMat, DoubleInv;
        LPGAMMATABLE InverseShapes[3];
@@ -487,8 +625,7 @@ LPMATSHAPER cmsBuildOutputMatrixShaper(cmsHPROFILE OutputProfile, LPDWORD dwFlag
        
 
        if (cmsGetColorSpace(OutputProfile) == icSigGrayData)
-       {
-              if (dwFlags) *dwFlags |= cmsFLAGS_NOTPRECALC;
+       {              
               return cmsBuildGrayOutputMatrixShaper(OutputProfile);              
        }
 
@@ -567,39 +704,18 @@ BOOL cmsBuildSmeltMatShaper(_LPcmsTRANSFORM p)
 
 // Conversion between PCS ------------------------------------------
 
-// Identifies intent archieved by LUT, only absolute colorimetric ones
-// are handled separatly
+// Identifies intent archieved by LUT
 
 static
 int GetPhase(cmsHPROFILE hProfile)
-{
-       icColorSpaceSignature PCS;
-       icRenderingIntent Intent;
+{       
+       switch (cmsGetPCS(hProfile)) {
 
-       PCS    = cmsGetPCS(hProfile);
-       Intent = (icRenderingIntent) cmsTakeRenderingIntent(hProfile);
+       case icSigXYZData: return XYZRel;
 
-       switch (PCS) {
-       case icSigXYZData:
-                    /*
-                     if (Intent == icAbsoluteColorimetric)
+       case icSigLabData: return LabRel;
 
-                            return XYZAbs;
-                     else
-                     */
-                            return XYZRel;
-
-       case icSigLabData:
-                     /*
-                     if (Intent == icAbsoluteColorimetric)
-
-                            return LabAbs;
-                     else
-                     */
-                            return LabRel;
-
-       default:
-                     cmsSignalError(LCMS_ERRC_ABORTED, "Invalid PCS");
+       default:  cmsSignalError(LCMS_ERRC_ABORTED, "Invalid PCS");
        }
 
        return XYZRel;
@@ -628,16 +744,54 @@ void TakeConversionRoutines(_LPcmsTRANSFORM p, int DoBPC)
        cmsTakeMediaBlackPoint(&BlackPointOut,  p -> OutputProfile);
        cmsReadChromaticAdaptationMatrix(&ChromaticAdaptationMatrixOut, p -> OutputProfile);
 
-       // Guess black point from intent table
-       if (DoBPC) {
+            
+       if (p -> Preview == NULL && p ->Gamut == NULL)     // Non-proofing
+       {            
+            if (p ->Intent == INTENT_PERCEPTUAL ||
+                p ->Intent == INTENT_SATURATION) {
+
+
+                    // For v4 profiles, Perceptual PCS has a reference black point
+                    // which v2 profiles should scale to.
+
+                    if ((cmsGetProfileICCversion(p ->InputProfile) >= 0x4000000) ||
+                        (cmsGetProfileICCversion(p ->OutputProfile) >= 0x4000000)) {
+
+                           DoBPC = TRUE;
+                    }
+            }
+
+            // Black point compensation does not apply to absolute intent
+
+            if (p ->Intent == INTENT_ABSOLUTE_COLORIMETRIC) 
+                            DoBPC = FALSE;
+
+            // Black point compensation does not apply to devicelink profiles
+            
+            if (cmsGetDeviceClass(p ->InputProfile) == icSigLinkClass)
+                            DoBPC = FALSE;
+
+            if (cmsGetDeviceClass(p ->OutputProfile) == icSigLinkClass)
+                            DoBPC = FALSE;
+            
+            if (DoBPC) {
        
-           cmsDetectBlackPoint(&BlackPointIn,    p->InputProfile,   p->Intent, 0);
-           cmsDetectBlackPoint(&BlackPointOut,   p->OutputProfile,  p->Intent, 0);          
-       }
-      
-       if (p -> Preview == NULL)     // Non-proofing
-       {
-              cmsChooseCnvrt(p -> Intent == INTENT_ABSOLUTE_COLORIMETRIC,
+                // Detect Black points
+
+                cmsDetectBlackPoint(&BlackPointIn,    p->InputProfile,   p->Intent, 0);
+                cmsDetectBlackPoint(&BlackPointOut,   p->OutputProfile,  p->Intent, 0);          
+
+                // If equal black points, then do nothing. This often applies to BP=0
+
+                if (BlackPointIn.X == BlackPointOut.X &&
+                    BlackPointIn.Y == BlackPointOut.Y &&
+                    BlackPointIn.Z == BlackPointOut.Z) 
+                                DoBPC = FALSE;
+
+
+            }
+            
+            cmsChooseCnvrt(p -> Intent == INTENT_ABSOLUTE_COLORIMETRIC,
 
                  p -> Phase1,
                              &BlackPointIn,
@@ -663,12 +817,26 @@ void TakeConversionRoutines(_LPcmsTRANSFORM p, int DoBPC)
        cmsTakeIluminant(&IlluminantProof,        p -> PreviewProfile);
        cmsTakeMediaWhitePoint(&WhitePointProof,  p -> PreviewProfile);
        cmsTakeMediaBlackPoint(&BlackPointProof,  p -> PreviewProfile);
-
-       if (DoBPC) 
-            cmsDetectBlackPoint(&BlackPointProof, p->PreviewProfile, p->Intent, 0);
-
        cmsReadChromaticAdaptationMatrix(&ChromaticAdaptationMatrixProof, p -> PreviewProfile);
 
+       if (DoBPC) {
+
+            cmsDetectBlackPoint(&BlackPointProof, p->PreviewProfile, p->Intent, 0);
+            cmsDetectBlackPoint(&BlackPointIn,    p->InputProfile,   p->Intent, 0);
+            cmsDetectBlackPoint(&BlackPointOut,   p->OutputProfile,  p->Intent, 0);  
+            
+            // Reality check
+
+            if (BlackPointIn.X == BlackPointProof.X &&
+                BlackPointIn.Y == BlackPointProof.Y &&
+                BlackPointIn.Z == BlackPointProof.Z) 
+                                DoBPC = FALSE;
+
+
+       }
+
+       
+
        cmsChooseCnvrt(p -> Intent == INTENT_ABSOLUTE_COLORIMETRIC,
 
                  p -> Phase1,
@@ -702,8 +870,6 @@ void TakeConversionRoutines(_LPcmsTRANSFORM p, int DoBPC)
                  0,
                  &p->Stage2,
                  &p->m2, &p->of2);
-
-
        }
 
 }
@@ -757,16 +923,47 @@ _LPcmsTRANSFORM AllocEmptyTransform(void)
     p -> OutMatShaper   = NULL;
     p -> SmeltMatShaper = NULL;
     p -> NamedColorList = NULL;
-    p -> EntryColorSpace = 0;
-    p -> ExitColorSpace  = 0;
+    p -> EntryColorSpace = (icColorSpaceSignature) 0;
+    p -> ExitColorSpace  = (icColorSpaceSignature) 0;
 
     return p;
 }
 
 
+// Identify whatever a transform is to be cached
+
+static
+void SetPrecalculatedTransform(_LPcmsTRANSFORM p, DWORD dwFlags)
+{
+    if (dwFlags & cmsFLAGS_GAMUTCHECK && p ->GamutCheck != NULL) {
+        
+        p -> xform = PrecalculatedXFORMGamutCheck;
+        
+        if (!(dwFlags & cmsFLAGS_NOTCACHE)) {
+            
+            ZeroMemory(p ->CacheIn, sizeof(WORD) * MAXCHANNELS);
+            TransformOnePixelWithGamutCheck(p, p->CacheIn, p ->CacheOut);        
+            p ->xform = CachedXFORMGamutCheck;              
+        }
+        
+    }
+    else {
+        
+        p -> xform = PrecalculatedXFORM;
+        
+        if (!(dwFlags & cmsFLAGS_NOTCACHE)) {
+            
+            ZeroMemory(p ->CacheIn, sizeof(WORD) * MAXCHANNELS);
+            cmsEvalLUT(p ->DeviceLink, p->CacheIn, p ->CacheOut);        
+            p ->xform = CachedXFORM;        
+        }
+    }
+}
+
+
 // Transform is identified as device-link
 static 
-cmsHPROFILE CreateDeviceLinkTransform(_LPcmsTRANSFORM p)
+cmsHPROFILE CreateDeviceLinkTransform(_LPcmsTRANSFORM p, DWORD dwFlags)
 {
     
     if (!IsProperColorSpace(p->InputProfile, p->InputFormat, FALSE)) {
@@ -801,8 +998,17 @@ cmsHPROFILE CreateDeviceLinkTransform(_LPcmsTRANSFORM p)
     p -> Phase1 = -1;
     p -> Phase2 = -1;
     p -> Phase3 = -1;
-    p -> xform = PrecalculatedXFORM;
+
+    SetPrecalculatedTransform(p, dwFlags);
+    
+    p -> EntryColorSpace = cmsGetColorSpace(p -> InputProfile);
     p -> ExitColorSpace = cmsGetPCS(p -> InputProfile);
+    
+    if (p ->EntryColorSpace == icSigRgbData ||
+        p ->EntryColorSpace == icSigCmyData) {
+                   
+                    p->DeviceLink -> CLut16params.Interp3D = cmsTetrahedralInterp16;
+    }
 
     // Precalculated device-link profile is ready
     return (cmsHTRANSFORM) p;
@@ -823,9 +1029,13 @@ void CreateProof(_LPcmsTRANSFORM p,
       // Apr-15, 2002 - Too much profiles does have bogus content
       // on preview tag, so I do compute it by my own.
 
-      p -> Preview = _cmsComputeSoftProofLUT(p ->PreviewProfile, p ->ProofIntent); 
+      p -> Preview = _cmsComputeSoftProofLUT(p ->PreviewProfile, p ->Intent); 
       p -> Phase2  = LabRel;
 
+      // That's a proofing transfor, so use second intent for output.
+
+      *ToTagPtr  = PCS2Device[p->ProofIntent];
+
       if (p -> Preview == NULL) {
           
         ProofTag = Preview[p -> Intent];
@@ -841,8 +1051,7 @@ void CreateProof(_LPcmsTRANSFORM p,
 
              p -> Preview = cmsReadICCLut(p ->PreviewProfile, ProofTag);
              p -> Phase2 = GetPhase(p ->PreviewProfile);
-             *ToTagPtr  = PCS2Device[p->ProofIntent];
-             
+                         
         }
         else
              {
@@ -858,10 +1067,12 @@ void CreateProof(_LPcmsTRANSFORM p,
     // Aug-31, 2001 - Too much profiles does have bogus content
     // on gamut tag, so I do compute it by my own.
 
-    if (dwFlags & cmsFLAGS_GAMUTCHECK) {
+    if ((dwFlags & cmsFLAGS_GAMUTCHECK) && (dwFlags & cmsFLAGS_NOTPRECALC)) {
 
                    
-             p -> Gamut = _cmsComputeGamutLUT(p->PreviewProfile, p ->ProofIntent);                               
+             p -> Gamut = _cmsComputeGamutLUT(p->PreviewProfile, p ->Intent);  
+             p -> Phase2  = LabRel;
+
              if (p -> Gamut == NULL) {
 
                 // Profile goes only in one direction... try to see 
@@ -871,6 +1082,7 @@ void CreateProof(_LPcmsTRANSFORM p,
                 if (cmsIsTag(p ->PreviewProfile, icSigGamutTag)) {
 
                     p -> Gamut = cmsReadICCLut(p ->PreviewProfile, icSigGamutTag);
+                    
                     }
                     else   {
                             
@@ -894,29 +1106,7 @@ _LPcmsTRANSFORM PickTransformRoutine(_LPcmsTRANSFORM p,
                                     icTagSignature *ToTagPtr)
 {
 
-        // Can we optimize matrix-shaper only transform?
-
-       if (*FromTagPtr == 0 && 
-           *ToTagPtr == 0 && 
-           !p->PreviewProfile  && 
-           p -> Intent != INTENT_ABSOLUTE_COLORIMETRIC && 
-           (p -> EntryColorSpace == icSigRgbData) && 
-           (p -> ExitColorSpace == icSigRgbData)) {
-
-              // Yes... try to smelt matrix-shapers
-              p -> xform = MatrixShaperXFORM;
-              *dwFlagsPtr |= cmsFLAGS_NOTPRECALC;
-
-              if (!cmsBuildSmeltMatShaper(p))
-              {
-                     cmsSignalError(LCMS_ERRC_ABORTED, "unable to smelt shaper-matrix, required tags missing");               
-                     return NULL;
-              }
-
-              p -> Phase1 = p -> Phase3 = XYZRel;
-              return p;
-
-       }
+       
        
 
         // Is a named color profile?
@@ -926,6 +1116,29 @@ _LPcmsTRANSFORM PickTransformRoutine(_LPcmsTRANSFORM p,
                   p ->FromDevice = NC2toPCS;
         }
         else {
+                // Can we optimize matrix-shaper only transform?
+
+                   if (*FromTagPtr == 0 && 
+                       *ToTagPtr == 0 && 
+                       !p->PreviewProfile  && 
+                       p -> Intent != INTENT_ABSOLUTE_COLORIMETRIC && 
+                       (p -> EntryColorSpace == icSigRgbData) && 
+                       (p -> ExitColorSpace == icSigRgbData)) {
+
+                          // Yes... try to smelt matrix-shapers
+                          p -> xform = MatrixShaperXFORM;
+                          *dwFlagsPtr |= cmsFLAGS_NOTPRECALC;
+
+                          if (!cmsBuildSmeltMatShaper(p))
+                          {
+                                 cmsSignalError(LCMS_ERRC_ABORTED, "unable to smelt shaper-matrix, required tags missing");               
+                                 return NULL;
+                          }
+
+                          p -> Phase1 = p -> Phase3 = XYZRel;
+                          return p;
+
+                }
 
                 // No, is a transform involving LUT
 
@@ -943,7 +1156,7 @@ _LPcmsTRANSFORM PickTransformRoutine(_LPcmsTRANSFORM p,
               else
               {
                      p -> FromDevice = ShaperMatrixToPCS;
-                     p -> InMatShaper = cmsBuildInputMatrixShaper(p -> InputProfile, dwFlagsPtr);
+                     p -> InMatShaper = cmsBuildInputMatrixShaper(p -> InputProfile);
 
                      if (!p ->InMatShaper) {
                             cmsSignalError(LCMS_ERRC_ABORTED, "profile is unsuitable for input");                            
@@ -968,7 +1181,7 @@ _LPcmsTRANSFORM PickTransformRoutine(_LPcmsTRANSFORM p,
               else
               {
                      p -> ToDevice = PCStoShaperMatrix;
-                     p -> OutMatShaper = cmsBuildOutputMatrixShaper(p->OutputProfile, dwFlagsPtr);
+                     p -> OutMatShaper = cmsBuildOutputMatrixShaper(p->OutputProfile);
 
                      if (!p -> OutMatShaper) {
                             cmsSignalError(LCMS_ERRC_ABORTED, "profile is unsuitable for output");                            
@@ -1002,10 +1215,10 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        icTagSignature ToTag;
        
        if (nIntent < 0 || nIntent > 3 ||
-           ProofingIntent < 0 || ProofingIntent > 3)
-       {
-       cmsSignalError(LCMS_ERRC_ABORTED, "cmsCreateTransform: intent mismatch");
-       return NULL;
+           ProofingIntent < 0 || ProofingIntent > 3) {
+
+            cmsSignalError(LCMS_ERRC_ABORTED, "cmsCreateTransform: intent mismatch");
+            return NULL;
        }
 
        p = AllocEmptyTransform();
@@ -1020,6 +1233,7 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        p -> PreviewProfile = ProofingProfile;            
        p -> InputFormat    = InputFormat;
        p -> OutputFormat   = OutputFormat;
+       p -> lInputV4Lab    = p ->lOutputV4Lab = FALSE;
 
        p -> FromInput = _cmsIdentifyInputFormat(p, InputFormat);
        p -> ToOutput  = _cmsIdentifyOutputFormat(p, OutputFormat);
@@ -1045,12 +1259,11 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        //  Device link are means to store precalculated transform grids.
        if (cmsGetDeviceClass(InputProfile) == icSigLinkClass) {
 
-            return CreateDeviceLinkTransform(p);
+            return CreateDeviceLinkTransform(p, dwFlags);
        }
-
-     
-     
+         
        if (!IsProperColorSpace(InputProfile, InputFormat, FALSE)) {
+
               cmsSignalError(LCMS_ERRC_ABORTED, "Input profile is operating on wrong colorspace");
               cmsDeleteTransform((cmsHTRANSFORM) p);
               return NULL;
@@ -1062,7 +1275,7 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        if (cmsGetDeviceClass(InputProfile) == icSigNamedColorClass) {
 
         if (p ->NamedColorList == NULL)
-                p ->NamedColorList = cmsAllocNamedColorList();
+                p ->NamedColorList = cmsAllocNamedColorList(0);
 
         cmsReadICCnamedColorList(p, InputProfile, icSigNamedColor2Tag);
         
@@ -1115,8 +1328,8 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        FromTag  = Device2PCS[nIntent];
        ToTag    = PCS2Device[nIntent];
 
-       if (!cmsIsTag(InputProfile, FromTag))
-       {
+       if (!cmsIsTag(InputProfile, FromTag)) {
+
               FromTag = Device2PCS[0];
 
               if (!cmsIsTag(InputProfile,  FromTag)) {
@@ -1129,7 +1342,6 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
            CreateProof(p, dwFlags, &ToTag);                       
        
 
-
        if (!cmsIsTag(OutputProfile,  ToTag)) {
 
            ToTag = PCS2Device[0];
@@ -1153,12 +1365,17 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
        if (dwFlags & cmsFLAGS_MATRIXOUTPUT)
               ToTag = (icTagSignature)0;
 
+       
+       // *** uncomment this to disable fast gamut checking 
+        /*
        if (dwFlags & cmsFLAGS_GAMUTCHECK) {
 
-              p -> DoGamutCheck = TRUE;
-              dwFlags |= cmsFLAGS_NOTPRECALC;
+             p -> DoGamutCheck = TRUE;
+             dwFlags |= cmsFLAGS_NOTPRECALC;
        }
+       */
 
+       // *** Up to here 
      
        if (PickTransformRoutine(p, &dwFlags, &FromTag, &ToTag) == NULL) {
 
@@ -1167,18 +1384,52 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
 
        }
 
-       TakeConversionRoutines(p, dwFlags & cmsFLAGS_WHITEBLACKCOMPENSATION);
+       TakeConversionRoutines(p, dwFlags & cmsFLAGS_BLACKPOINTCOMPENSATION);
 
      
        if (!(dwFlags & cmsFLAGS_NOTPRECALC)) {
 
-               LPLUT DeviceLink;    
-
+               LPLUT DeviceLink;   
+               LPLUT GamutCheck = NULL;
+               
                DeviceLink = _cmsPrecalculateDeviceLink((cmsHTRANSFORM) p, dwFlags);
-                                        
+                
+               if (dwFlags & cmsFLAGS_GAMUTCHECK) {
+
+                   GamutCheck = _cmsPrecalculateGamutCheck((cmsHTRANSFORM) p);
+               }
+
+               // If input colorspace is Rgb, Cmy, then use tetrahedral interpolation
+               // for speed reasons (it only works well on spaces on Luma is diagonal, and 
+               // not if luma is in separate channel)
+
+               if (p ->EntryColorSpace == icSigRgbData ||
+                   p ->EntryColorSpace == icSigCmyData) {
+                   
+                    DeviceLink -> CLut16params.Interp3D = cmsTetrahedralInterp16;
+               }
+               
+               // If this is a 8-bit transform, optimize LUT further.
+                                           
+                if ((T_BYTES(InputFormat) == 1) && (T_CHANNELS(InputFormat) == 3)) {
+
+                    DeviceLink = _cmsBlessLUT8(DeviceLink);
+                    if (DeviceLink == NULL) return NULL;
+
+                }
+                
+
+                p ->GamutCheck = GamutCheck;
+
                 if (DeviceLink) {
 
                     p ->DeviceLink = DeviceLink;
+
+                    if ((nIntent != INTENT_ABSOLUTE_COLORIMETRIC) && 
+                        !(dwFlags & cmsFLAGS_NOWHITEONWHITEFIXUP))
+
+							_cmsFixWhiteMisalignment(p);
+
                 }
                 else
                 {
@@ -1192,7 +1443,8 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateProofingTransform(cmsHPROFILE InputProfile,
                 }
 
 
-              p -> xform = PrecalculatedXFORM;
+              SetPrecalculatedTransform(p, dwFlags);
+              
 
        }
 
@@ -1339,7 +1591,7 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateMultiprofileTransform(cmsHPROFILE hProfiles[],
                                                                 DWORD dwFlags)
 {
     cmsHTRANSFORM Transforms[257];
-    DWORD dwPrecalcFlags = (dwFlags|cmsFLAGS_NOTPRECALC);
+    DWORD dwPrecalcFlags = (dwFlags|cmsFLAGS_NOTPRECALC|cmsFLAGS_NOTCACHE);
     DWORD FormatInput, FormatOutput;
     cmsHPROFILE hLab, hXYZ, hProfile;   
     icColorSpaceSignature ColorSpace, CurrentColorSpace;
@@ -1355,7 +1607,8 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateMultiprofileTransform(cmsHPROFILE hProfiles[],
     }
 
     // Creates a phantom transform for latter filling
-    p = cmsCreateTransform(NULL, dwInput, NULL, dwOutput, Intent, cmsFLAGS_NULLTRANSFORM);
+    p = (_LPcmsTRANSFORM) cmsCreateTransform(NULL, dwInput, 
+                                             NULL, dwOutput, Intent, cmsFLAGS_NULLTRANSFORM);
 
     // If user wants null one, give it
     if (dwFlags & cmsFLAGS_NULLTRANSFORM) return (cmsHPROFILE) p;
@@ -1373,13 +1626,23 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateMultiprofileTransform(cmsHPROFILE hProfiles[],
             // Yes, only named color. Create a named color-device
             // and append to named color table
 
-            p = cmsCreateTransform(hProfiles[0], dwInput, NULL, dwOutput, Intent, dwFlags);
+            cmsDeleteTransform((cmsHTRANSFORM) p);
+
+            p = (_LPcmsTRANSFORM) cmsCreateTransform(hProfiles[0], dwInput, NULL, dwOutput, Intent, dwFlags);
             for (i=1; i < nProfiles; i++) {
                     cmsReadICCnamedColorList(p, hProfiles[i], icSigNamedColor2Tag);
             }
 
             return p;   // Ok, done so far
     }
+    else
+    if (nNamedColor > 0) {
+
+        cmsDeleteTransform((cmsHTRANSFORM) p);
+        cmsSignalError(LCMS_ERRC_ABORTED, "Could not mix named color profiles with other types in multiprofile transform");
+        return NULL;
+    }
+
 
     // We will need a 3DCLUT for device link
     Grid =  cmsAllocLUT();
@@ -1486,17 +1749,14 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateMultiprofileTransform(cmsHPROFILE hProfiles[],
     p ->InputProfile  = hProfiles[0];
     p ->OutputProfile = hProfiles[nProfiles - 1];
 
-
-
     nGridPoints = _cmsReasonableGridpointsByColorspace(p ->EntryColorSpace, dwFlags);
    
     ChannelsInput  = _cmsChannelsOf(cmsGetColorSpace(p ->InputProfile));
             
     Grid = cmsAlloc3DGrid(Grid, nGridPoints, ChannelsInput, ChannelsOutput);
 
-    if ((p ->EntryColorSpace == icSigRgbData) && (p->ExitColorSpace == icSigRgbData)
-        &&  !(dwFlags & cmsFLAGS_NOPRELINEARIZATION))
-                    _cmsComputePrelinearizationTablesFromXFORM(Transforms, nProfiles, Grid);
+    if (!(dwFlags & cmsFLAGS_NOPRELINEARIZATION))
+           _cmsComputePrelinearizationTablesFromXFORM(Transforms, nProfiles, Grid);
       
     // Compute device link on 16-bit basis                
     if (!cmsSample3DGrid(Grid, MultiprofileSampler, (LPVOID) Transforms, Grid -> wFlags)) {
@@ -1507,15 +1767,20 @@ cmsHTRANSFORM LCMSEXPORT cmsCreateMultiprofileTransform(cmsHPROFILE hProfiles[],
 
     // All ok, store the newly created LUT   
     p -> DeviceLink   = Grid;
-    p -> xform        = PrecalculatedXFORM;
-
 
+    SetPrecalculatedTransform(p, dwFlags);
+    
     for (i=nProfiles-1; i >= 0; --i)
         cmsDeleteTransform(Transforms[i]);
 
 
     if (hLab) cmsCloseProfile(hLab);
     if (hXYZ) cmsCloseProfile(hXYZ);
+
+    if ((Intent != INTENT_ABSOLUTE_COLORIMETRIC) && 
+        !(dwFlags & cmsFLAGS_NOWHITEONWHITEFIXUP))
+							_cmsFixWhiteMisalignment(p);
+
     return (cmsHTRANSFORM) p;
 
 
